/*
 * Copyright @ 1984 - 1993   Josef Heinen
 *
 * Permission to use, copy, and distribute this software and its
 * documentation for any purpose with or without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear
 * in supporting documentation.
 *
 * Permission to modify the software is granted, but not the right to
 * distribute the modified code.  Modifications are to be distributed
 * as patches to released version.
 *
 * This software is provided "as is" without express or implied warranty.
 *
 * Send your comments or suggestions to
 *  J.Heinen@kfa-juelich.de.
 *
 */

#include <stdio.h>

/* Simple types */
 
typedef FILE            Gfile;
typedef char            Gchar;
typedef char            Gconn;
typedef float           Gfloat ;
typedef int             Gwstype;
typedef int             Gint;
typedef unsigned int    Guint;
typedef long            Glong;

#define GWC_DEF         NULL
#define GCONID_DEFAULT  NULL

#define GWS_DEFAULT     NULL		/* Default Workstation type */
#define GWS_DEF         NULL

/* Enumerations */

typedef enum {          /* aspect control flag */
        GBUNDLED,
        GINDIVIDUAL
} Gasf;

typedef enum {          /* clipping indicator */
        GCLIP,
        GNOCLIP
} Gclip;

typedef enum {          /* clear control flag */
        GCONDITIONALLY,
        GALWAYS
} Gclrflag;

typedef enum {          /* co-ord switch */
        GWC,
        GNDC
} Gcsw;

typedef enum {          /* device co-ord units */
        GDC_METRES,
        GDC_OTHER
} Gdevunits;

typedef enum {          /* fill area interior style */
        GHOLLOW,
        GSOLID,
        GPATTERN,
        GHATCH
} Gflinter;

typedef enum {          /* request status */
        GOK,
        GNONE
} Gistat;

typedef enum {          /* line type - not standard */
        GLN_SOLID = 1,
        GLN_DASHED,
        GLN_DOTTED,
        GLN_DASHDOT,
        GLN_TRIPLE_DOT = -8,
        GLN_DOUBLE_DOT,
        GLN_SPACED_DOT,
        GLN_SPACED_DASH,
        GLN_LONG_SHORT_DASH,
        GLN_LONG_DASH,
        GLN_DASH_3_DOT,
        GLN_DASH_2_DOT
} Glntype;

typedef enum {          /* marker type - not standard */
        GMK_POINT = 1,
        GMK_PLUS,
        GMK_STAR,
        GMK_O,
        GMK_X,
        GMK_SOLID_DIAMOND = -13,
        GMK_DIAMOND,
        GMK_SOLID_HGLASS,
        GMK_HOURGLASS,
        GMK_SOLID_BOWTIE,
        GMK_BOWTIE,
        GMK_SOLID_SQUARE,
        GMK_SQUARE,
        GMK_SOLID_TRI_DOWN,
        GMK_TRIANGLE_DOWN,
        GMK_SOLID_TRI_UP,
        GMK_TRIANGLE_UP,
        GMK_SOLID_CIRCLE
} Gmktype;

typedef enum {          /* GKS operating state */
        GGKCL,          /* closed */
        GGKOP,          /* open */
        GWSOP,          /* workstation open */
        GWSAC,          /* workstation active */
        GSGOP           /* segment open */
} Gopst;

typedef enum {          /* regeneration flag */
        GPERFORM,
        GPOSTPONE
} Gregen;

typedef enum {          /* horiz text alignment component */
        GAH_NORMAL,
        GAH_LEFT,
        GAH_CENTRE,
        GAH_RIGHT
} Gtxhor;

typedef enum {          /* text path */
        GTP_RIGHT,
        GTP_LEFT,
        GTP_UP,
        GTP_DOWN
} Gtxpath;

typedef enum {          /* text precision */
        GP_STRING,
        GP_CHAR,
        GP_STROKE
} Gtxprec;

typedef enum {          /* vert text alignment component */
        GAV_NORMAL,
        GAV_TOP,
        GAV_CAP,
        GAV_HALF,
        GAV_BASE,
        GAV_BOTTOM
} Gtxver;

typedef enum {          /* WS category */
        GOUTPUT,
        GINPUT,
        GOUTIN,
        GWISS,
        GMO,
        GMI
} Gwscat;
 
/* Forward type definitions */

typedef struct {        /* integer point */
        Gint    x;              /* x coordinate */
        Gint    y;              /* y coordinate */
} Gipoint;

typedef struct {        /* coordinate point */
        Gfloat  x;              /* X coordinate */
        Gfloat  y;              /* Y coordinate */
} Gpoint;

typedef struct {        /* coordinate limits */
        Gfloat  xmin;           /* x minimum limit */
        Gfloat  xmax;           /* x maximum limit */
        Gfloat  ymin;           /* y minimum limit */
        Gfloat  ymax;           /* y maximum limit */
} Glimit;

typedef struct {        /* text facilities */
        Gint    font;           /* text font */
        Gtxprec prec;           /* text precision */
} Gtxfp;

typedef struct {        /* text alignment */
        Gtxhor  hor;            /* horizontal component */
        Gtxver  ver;            /* vertical component */
} Gtxalign;

typedef struct {        /* coordinate rectangle pointer */
        Gpoint  ul;             /* upper left point */
        Gpoint  lr;             /* lower right point */
} Grect;

typedef struct {        /* dimension in integer values */
        Guint   x_dim;          /* X dimension */
        Guint   y_dim;          /* Y dimension */
} Gidim;

/* Structs */
 
typedef struct {        /* aspect source flags */
        Gasf    ln_type;        /* line type */
        Gasf    ln_width;       /* line width */
        Gasf    ln_colour;      /* line colour */
        Gasf    mk_type;        /* marker type */
        Gasf    mk_size;        /* marker size */
        Gasf    mk_colour;      /* marker colour */
        Gasf    tx_fp;          /* text font and precision */
        Gasf    tx_exp;         /* text expansion */
        Gasf    tx_space;       /* text character spacing */
        Gasf    tx_colour;      /* text colour */
        Gasf    fl_inter;       /* fill area interior style */
        Gasf    fl_style;       /* fill area style index */
        Gasf    fl_colour;      /* fill area colour */
} Gasfs;

typedef struct {        /* colour bundle */
        Gfloat  red;            /* red intensity */
        Gfloat  green;          /* green intensity */
        Gfloat  blue;           /* blue intensity */
} Gcobundl;

typedef struct {        /* clipping rectangle */
        Gclip   ind;    /* clipping indicator */
        Glimit  rec;    /* clipping rectangle */
} Gcliprect;

typedef struct {        /* display size */
        Gdevunits units;        /* device coordinate units */
        Gpoint  device;         /* device coordinate unit size */
        Gipoint raster;         /* raster unit size */
} Gdspsize;

typedef struct {        /* text extent */
        Gpoint  concat;         /* concatenation point */
        Gpoint  corner_1;       /* corner 1 */
        Gpoint  corner_2;       /* corner 2 */
        Gpoint  corner_3;       /* corner 3 */
        Gpoint  corner_4;       /* corner 4 */
} Gextent;

typedef struct {        /* locator data */
        Gint    transform;      /* normalization transformation number */
        Gpoint  position;       /* locator position */
} Gloc;

typedef struct {        /* request locator */
        Gistat  status;         /* request status */
        Gloc    loc;            /* locator data */
} Gqloc;

typedef struct {        /* request string */
        Gistat  status;         /* request status */
        Gchar   *string;        /* string data */
} Gqstring;
 
typedef struct {        /* scale vector */
        Gfloat x_scale;
        Gfloat y_scale;
} Gscale;

typedef struct {        /* transformation */
        Glimit  w;              /* window */
        Glimit  v;              /* viewport */
} Gtran;

typedef struct {        /* metafile item information */
        Gint    type;           /* item type */
        Gint    length;         /* item length */
} Ggksmit;

#define gsetlinecolorind gsetlinecolourind
#define gsetmarkercolorind gsetmarkercolourind
#define gsettextcolorind gsettextcolourind
#define gsetfillcolorind gsetfillcolourind
#define gsetcolorrep gsetcolourrep
#define ginqlinecolorind ginqlinecolourind
#define ginqmarkercolorind ginqmarkercolourind
#define ginqtextcolorind ginqtextcolourind
#define ginqfillcolorind ginqfillcolourind
