/*
 * Copyright @ 1984 - 1993   Josef Heinen
 *
 * Permission to use, copy, and distribute this software and its
 * documentation for any purpose with or without fee is hereby granted, 
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear
 * in supporting documentation.
 *
 * Permission to modify the software is granted, but not the right to
 * distribute the modified code.  Modifications are to be distributed
 * as patches to released version.
 *
 * This software is provided "as is" without express or implied warranty.
 *
 * Send your comments or suggestions to
 *  J.Heinen@kfa-juelich.de.
 *
 *
 * FACILITY:
 *
 *      GLI GKS V4.5
 *
 * ABSTRACT:
 *
 *      This module contains the C language binding for GLI GKS.
 *
 * AUTHOR:
 *
 *      Josef Heinen
 *
 * VERSION:
 *
 *      V1.0
 *
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#ifndef MSDOS
#include <sys/types.h>
#endif

#if defined (cray) || defined (__SVR4) || defined(MSDOS)
#include <fcntl.h>
#else
#include <sys/file.h>
#endif

#include "gks.h"
 
#ifdef VMS
#include <descrip.h>
#endif

#ifdef cray
#include <fortran.h>
#endif

#define OK          0
#define MAX_POINTS  2048


/* Entry point definitions */

#ifndef cray
#if defined (VMS) || defined (hpux) || defined (aix)

#define GOPKS	gopks
#define GCLKS	gclks
#define GOPWK	gopwk
#define GRSGWK  grsgwk
#define GCLWK	gclwk
#define GACWK	gacwk
#define GDAWK	gdawk
#define GCLRWK	gclrwk
#define GUWK	guwk
#define GECLKS	geclks
#define GESC	gesc
#define GMSG	gmsg
#define GMSGS	gmsgs

#define GPL	gpl
#define GPM	gpm
#define GTX	gtx
#define GTXS	gtxs
#define GFA	gfa
#define GCA	gca

#define GSASF	gsasf
#define GSPLI	gspli
#define GSLN	gsln
#define GSLWSC	gslwsc
#define GSPLCI	gsplci
#define GSPMI	gspmi
#define GSMK	gsmk
#define GSMKSC	gsmksc
#define GSPMCI	gspmci
#define GSCR	gscr
#define GSTXI	gstxi
#define GSTXFP	gstxfp
#define GSCHXP	gschxp
#define GSCHCI	gschci
#define GSCHH	gschh
#define GSCHUP	gschup
#define GSTXP	gstxp
#define GSCHSP	gschsp
#define GSTXAL	gstxal
#define GSTXCI	gstxci
#define GSFAI	gsfai
#define GSFAIS	gsfais
#define GSFASI	gsfasi
#define GSFACI	gsfaci

#define GSWN	gswn
#define GSVP	gsvp
#define GSELNT	gselnt
#define GSCLIP	gsclip
#define GSWKWN	gswkwn
#define GSWKVP	gswkvp

#define GCRSG   gcrsg
#define GCLSG   gclsg
#define GCSGWK  gcsgwk
#define GSSGT   gssgt
#define GSDS    gsds
#define GEVTM   gevtm

#define GINLC	ginlc
#define GRQLC	grqlc
#define GRQSK	grqsk
#define GRQST	grqst

#define GQOPS	gqops
#define GQLVKS	gqlvks
#define GQEWK	gqewk
#define GQMNTN	gqmntn
#define GQOPWK	gqopwk
#define GQACWK	gqacwk

#define GQASF	gqasf
#define GQPLI	gqpli
#define GQLN	gqln
#define GQLWSC	gqlwsc
#define GQPLCI	gqplci
#define GQPMI	gqpmi
#define GQMK	gqmk
#define GQMKSC	gqmksc
#define GQPMCI	gqpmci
#define GQTXI	gqtxi
#define GQTXFP	gqtxfp
#define GQCHXP	gqchxp
#define GQCHSP	gqchsp
#define GQTXCI	gqtxci
#define GQCHH	gqchh
#define GQCHUP	gqchup
#define GQTXP	gqtxp
#define GQTXAL	gqtxal
#define GQFAI	gqfai
#define GQFAIS	gqfais
#define GQFASI	gqfasi
#define GQFACI	gqfaci

#define GQCNTN	gqcntn
#define GQNT	gqnt
#define GQCLIP	gqclip
#define GQWKC	gqwkc
#define GQWKCA	gqwkca
#define GQTXX	gqtxx
#define GQTXXS	gqtxxs
#define GQDSP	gqdsp
#define GQOPSG  gqopsg

#else

#define GOPKS	gopks_
#define GCLKS	gclks_
#define GOPWK	gopwk_
#define GRSGWK  grsgwk_
#define GCLWK	gclwk_
#define GACWK	gacwk_
#define GDAWK	gdawk_
#define GCLRWK	gclrwk_
#define GUWK	guwk_
#define GECLKS	geclks_
#define GESC	gesc_
#define GMSG	gmsg_
#define GMSGS	gmsgs_

#define GPL	gpl_
#define GPM	gpm_
#define GTX	gtx_
#define GTXS	gtxs_
#define GFA	gfa_
#define GCA	gca_

#define GSASF	gsasf_
#define GSPLI	gspli_
#define GSLN	gsln_
#define GSLWSC	gslwsc_
#define GSPLCI	gsplci_
#define GSPMI	gspmi_
#define GSMK	gsmk_
#define GSMKSC	gsmksc_
#define GSPMCI	gspmci_
#define GSCR	gscr_
#define GSTXI	gstxi_
#define GSTXFP	gstxfp_
#define GSCHXP	gschxp_
#define GSCHCI	gschci_
#define GSCHH	gschh_
#define GSCHUP	gschup_
#define GSTXP	gstxp_
#define GSCHSP	gschsp_
#define GSTXAL	gstxal_
#define GSTXCI	gstxci_
#define GSFAI	gsfai_
#define GSFAIS	gsfais_
#define GSFASI	gsfasi_
#define GSFACI	gsfaci_

#define GSWN	gswn_
#define GSVP	gsvp_
#define GSELNT	gselnt_
#define GSCLIP	gsclip_
#define GSWKWN	gswkwn_
#define GSWKVP	gswkvp_

#define GCRSG   gcrsg_
#define GCLSG   gclsg_
#define GCSGWK  gcsgwk_
#define GSSGT   gssgt_
#define GSDS    gsds_
#define GEVTM   gevtm_

#define GINLC	ginlc_
#define GRQLC	grqlc_
#define GRQSK	grqsk_
#define GRQST	grqst_

#define GQOPS	gqops_
#define GQLVKS	gqlvks_
#define GQEWK	gqewk_
#define GQMNTN	gqmntn_
#define GQOPWK	gqopwk_
#define GQACWK	gqacwk_

#define GQASF	gqasf_
#define GQPLI	gqpli_
#define GQLN	gqln_
#define GQLWSC	gqlwsc_
#define GQPLCI	gqplci_
#define GQPMI	gqpmi_
#define GQMK	gqmk_
#define GQMKSC	gqmksc_
#define GQPMCI	gqpmci_
#define GQTXFP	gqtxfp_
#define GQCHXP	gqchxp_
#define GQCHSP	gqchsp_
#define GQTXI	gqtxi_
#define GQTXCI	gqtxci_
#define GQCHH	gqchh_
#define GQCHUP	gqchup_
#define GQTXP	gqtxp_
#define GQTXAL	gqtxal_
#define GQFAI	gqfai_
#define GQFAIS	gqfais_
#define GQFASI	gqfasi_
#define GQFACI	gqfaci_

#define GQCNTN	gqcntn_
#define GQNT	gqnt_
#define GQCLIP	gqclip_
#define GQWKC	gqwkc_
#define GQWKCA	gqwkca_
#define GQTXX	gqtxx_
#define GQTXXS	gqtxxs_
#define GQDSP	gqdsp_
#define GQOPSG  gqopsg_

#endif
#endif /* cray */

#ifdef VAXC
noshare extern int gks_errno;
#else
extern int gks_errno;
#endif

static float *x = NULL, *y = NULL;
static int max_points = 0;


static
gksrealloc (n)
    int n;
{
    if (n > max_points)
    {
        x = (float *) realloc (x, sizeof(float) * n);
        y = (float *) realloc (y, sizeof(float) * n);
        max_points = n;
    }
}


gopengks (errfile, memory)
    Gfile *errfile;
    Glong memory;
{
    int errfil, buffer = memory;

    errfil = (errfile != NULL) ? fileno(errfile) : 0;

    GOPKS (&errfil, &buffer);

    if (gks_errno == 0)
    {
        x = (float *) malloc (sizeof(float) * MAX_POINTS);
        y = (float *) malloc (sizeof(float) * MAX_POINTS);
        max_points = MAX_POINTS;
    }

    return gks_errno;
}


gclosegks ()
{
    GCLKS ();

    if (gks_errno == 0)
    {
        free (y);
        free (x);
        max_points = 0;
    }

    return gks_errno;
}
 

gopenws (workstation_id, connection, type)
    Gint workstation_id;
    Gconn *connection;
    Gwstype *type;
{
    int wkid = workstation_id, conid, wstype;

    conid = (connection != NULL) ?
        conid = open (connection, O_CREAT | O_TRUNC | O_WRONLY, 0644) : 0;
    wstype = (type != NULL) ? *type : 0;

    GOPWK (&wkid, &conid, &wstype);

    return gks_errno;
}
 

gclosews (workstation_id)
    Gint workstation_id;
{
    int wkid = workstation_id, errind, conid = 0, wtype;

    GQWKC (&wkid, &errind, &conid, &wtype);
    GCLWK (&wkid);

    if (conid > 0)
        close (conid);

    return gks_errno;
}
 

gactivatews (workstation_id)
    Gint workstation_id;
{
    int wkid = workstation_id;

    GACWK (&wkid);

    return gks_errno;
}

 
gdeactivatews (workstation_id)
    Gint workstation_id;
{
    int wkid = workstation_id;

    GDAWK (&wkid);

    return gks_errno;
}
 

gclearws (workstation_id, clearflag)
    Gint workstation_id;
    Gclrflag clearflag;
{
    int wkid = workstation_id, cofl = clearflag;

    GCLRWK (&wkid, &cofl);

    return gks_errno;
}
 

gupdatews (workstation_id, regenflag)
    Gint workstation_id;
    Gregen regenflag;
{
    int wkid = workstation_id, refl = regenflag;

    GUWK (&wkid, &refl);

    return gks_errno;
}
 

/*
gescape ()
{
} */


gmessage (workstation_id, string)
    Gint workstation_id;
    Gchar *string;
{
    int wkid = workstation_id, nchars;
    char *chars;
#ifdef VMS
    struct dsc$descriptor_s text;
#endif
#ifdef cray
    _fcd text;
#endif

    chars = string;
    nchars = strlen(chars);
#ifdef VMS
    text.dsc$b_dtype = DSC$K_DTYPE_T;
    text.dsc$b_class = DSC$K_CLASS_S;
    text.dsc$w_length = nchars;
    text.dsc$a_pointer = chars;

    GMSGS (&wkid, &nchars, &text);
#else
#ifdef cray
    text = _cptofcd(chars, nchars);

    GMSGS (&wkid, &nchars, text);
#else
    GMSGS (&wkid, &nchars, chars, nchars);
#endif /* cray */
#endif /* VMS */

    return gks_errno;
}


gpolyline (n, points)
    Gint n;
    Gpoint *points;
{
    int i, npoints = n;
 
    gksrealloc (n);

    for (i = 0; i < n; i++)
    {
        x[i] = points[i].x;
        y[i] = points[i].y;
    }
    GPL (&npoints, x, y);

    return gks_errno;
}
 

gpolymarker (n, points)
    Gint n;
    Gpoint *points;
{
    int i, npoints = n;
 
    gksrealloc (n);

    for (i = 0; i < n; i++)
    {
        x[i] = points[i].x;
        y[i] = points[i].y;
    }
    GPM (&npoints, x, y);

    return gks_errno;
}
 

gtext (position, string)
    Gpoint *position;
    Gchar *string;
{
    float qx, qy;
    int nchars;
    char *chars;
#ifdef VMS
    struct dsc$descriptor_s text;
#endif
#ifdef cray
    _fcd text;
#endif

    qx = position->x;
    qy = position->y;
    chars = string;
    nchars = strlen(chars);
#ifdef VMS
    text.dsc$b_dtype = DSC$K_DTYPE_T;
    text.dsc$b_class = DSC$K_CLASS_S;
    text.dsc$w_length = nchars;
    text.dsc$a_pointer = chars;

    GTXS (&qx, &qy, &nchars, &text);
#else
#ifdef cray
    text = _cptofcd(chars, nchars);

    GTXS (&qx, &qy, &nchars, text);
#else
    GTXS (&qx, &qy, &nchars, chars, nchars);
#endif /* cray */
#endif /* VMS */

    return gks_errno;
}
 

gfillarea (n, points)
    Gint n;
    Gpoint *points;
{
    int i, npoints = n;
 
    gksrealloc (n);
 
    for (i = 0; i < n; i++)
    {
        x[i] = points[i].x;
        y[i] = points[i].y;
    }
    GFA (&npoints, x, y);

    return gks_errno;
}
 

gcellarray (rectangle, dimensions, color)
    Grect *rectangle;
    Gidim *dimensions;
    Gint *color;
{
    float qx, qy, rx, ry;
    int dx, dy, scol, srow, ncol, nrow, *colia = color;

    qx = rectangle->ul.x;
    qy = rectangle->ul.y;
    rx = rectangle->lr.x;
    ry = rectangle->lr.y;
    dx = dimensions->x_dim;
    dy = dimensions->y_dim;
    scol = 1;
    srow = 1;
    ncol = dx;
    nrow = dy;

    GCA (&qx, &qy, &rx, &ry, &dx, &dy, &scol, &srow, &ncol, &nrow, colia);

    return gks_errno;
}


gsetasf (asfs)
    Gasfs *asfs;
{
    int flag[13];

    flag[0]  = asfs->ln_type;
    flag[1]  = asfs->ln_width;
    flag[2]  = asfs->ln_colour;
    flag[3]  = asfs->mk_type;
    flag[4]  = asfs->mk_size;
    flag[5]  = asfs->mk_colour;
    flag[6]  = asfs->tx_fp;
    flag[7]  = asfs->tx_exp;
    flag[8]  = asfs->tx_space;
    flag[9]  = asfs->tx_colour;
    flag[10] = asfs->fl_inter;
    flag[11] = asfs->fl_style;
    flag[12] = asfs->fl_colour;

    GSASF (flag);

    return gks_errno;
}
 

gsetlineind (index)
    Gint index;
{
    int pli = index;

    GSPLI (&pli);

    return gks_errno;
}


gsetlinetype (type)
    Gint type;
{
    int ltype = type;

    GSLN (&ltype);

    return gks_errno;
}
 

gsetlinewidth (width)
    Gfloat width;
{
    float lwidth = width;

    GSLWSC (&lwidth);

    return gks_errno;
}
 

gsetlinecolourind (colour)
    Gint colour;
{
    int coli = colour;

    GSPLCI (&coli);

    return gks_errno;
}
 

gsetmarkerind (index)
    Gint index;
{
    int pmi = index;

    GSPMI (&pmi);

    return gks_errno;
}


gsetmarkertype (type)
    Gint type;
{
    int mtype = type;

    GSMK (&mtype);

    return gks_errno;
}
 

gsetmarksize (size)
    Gfloat size;
{
    float mszsc = size;

    GSMKSC (&mszsc);

    return gks_errno;
}
 

gsetmarkercolourind (colour)
    Gint colour;
{
    int coli = colour;

    GSPMCI (&coli);

    return gks_errno;
}
 

gsettextind (index)
    Gint index;
{
    int txi = index;

    GSTXI (&txi);

    return gks_errno;
}


gsettextfontprec (txfp)
    Gtxfp *txfp;
{
    int font = txfp->font, prec = txfp->prec;

    GSTXFP (&font, &prec);

    return gks_errno;
}
 

gsetcharexpan (exp)
    Gfloat exp;
{
    float chxp = exp;

    GSCHXP (&chxp);

    return gks_errno;
}
 

gsetcharspace (spacing)
    Gfloat spacing;
{
    float chsp = spacing;

    GSCHSP (&chsp);

    return gks_errno;
}
 

gsettextcolourind (colour)
    Gint colour;
{
    int coli = colour;

    GSTXCI (&coli);

    return gks_errno;
}
 

gsetcharheight (height)
    Gfloat height;
{
    float chh = height;

    GSCHH (&chh);

    return gks_errno;
}
 

gsetcharup (charup)
    Gpoint *charup;
{
    float chux, chuy;

    chux = charup->x;
    chuy = charup->y;

    GSCHUP (&chux, &chuy);

    return gks_errno;
}
 

gsettextpath (text_path)
    Gtxpath text_path;
{
    int txp = text_path;

    GSTXP (&txp);

    return gks_errno;
}
 

gsettextalign (txalign)
    Gtxalign *txalign;
{
    int alh, alv;

    alh = txalign->hor;
    alv = txalign->ver;

    GSTXAL (&alh, &alv);

    return gks_errno;
}
 

gsetfillind (index)
    Gint index;
{
    int fai = index;

    GSFAI (&fai);

    return gks_errno;
}


gsetfillintstyle (style)
    Gflinter style;
{
    int ints = style;

    GSFAIS (&ints);

    return gks_errno;
}
 

gsetfillstyle (index)
    Gint index;
{
    int styli = index;

    GSFASI (&styli);

    return gks_errno;
}
 

gsetfillcolourind (colour)
    Gint colour;
{
    int coli = colour;

    GSFACI (&coli);

    return gks_errno;
}
 

gsetcolourrep (workstation_id, index, rep)
    Gint workstation_id, index;
    Gcobundl *rep;
{
    int wkid = workstation_id, coli = index;
    float r, g, b;

    r = rep->red;
    g = rep->green;
    b = rep->blue;

    GSCR (&wkid, &coli, &r, &g, &b);

    return gks_errno;
}
 

gsetwindow (transform, window)
    Gint transform;
    Glimit *window;
{
    int tnr = transform;
    float xmin, xmax, ymin, ymax;

    xmin = window->xmin;
    xmax = window->xmax;
    ymin = window->ymin;
    ymax = window->ymax;

    GSWN (&tnr, &xmin, &xmax, &ymin, &ymax);

    return gks_errno;
}
 

gsetviewport (transform, viewport)
    Gint transform;
    Glimit *viewport;
{
    int tnr = transform;
    float xmin, xmax, ymin, ymax;

    xmin = viewport->xmin;
    xmax = viewport->xmax;
    ymin = viewport->ymin;
    ymax = viewport->ymax;

    GSVP (&tnr, &xmin, &xmax, &ymin, &ymax);

    return gks_errno;
}

 
gselntran (transform)
    Gint transform;
{
    int tnr = transform;

    GSELNT (&tnr);

    return gks_errno;
}
 

gsetclip (indicator)
    Gclip indicator;
{
    int clsw = indicator;

    GSCLIP (&clsw);

    return gks_errno;
}
 

gsetwswindow (workstation_id, window)
    Gint workstation_id;
    Glimit *window;
{
    int wkid = workstation_id;
    float xmin, xmax, ymin, ymax;

    xmin = window->xmin;
    xmax = window->xmax;
    ymin = window->ymin;
    ymax = window->ymax;

    GSWKWN (&wkid, &xmin, &xmax, &ymin, &ymax);

    return gks_errno;
}

 
gsetwsviewport (workstation_id, viewport)
    Gint workstation_id;
    Glimit *viewport;
{
    int wkid = workstation_id;
    float xmin, xmax, ymin, ymax;

    xmin = viewport->xmin;
    xmax = viewport->xmax;
    ymin = viewport->ymin;
    ymax = viewport->ymax;

    GSWKVP (&wkid, &xmin, &xmax, &ymin, &ymax);

    return gks_errno;
}
 

/*
ginitloc ()
{
} */


greqloc (workstation_id, device_number, response)
    Gint workstation_id, device_number;
    Gqloc *response;
{
    int wkid = workstation_id, lcdnr = device_number;
    int stat, tnr;
    float qx, qy;

    GRQLC (&wkid, &lcdnr, &stat, &tnr, &qx, &qy);

    response->status = stat;
    response->loc.transform = tnr;
    response->loc.position.x = qx;
    response->loc.position.y = qy;

    return gks_errno;
}
 

/*
greqstroke ()
{
} */


greqstring (workstation_id, device_number, response)
    Gint workstation_id, device_number;
    Gqstring *response;
{
    int wkid = workstation_id, stdnr = device_number;
    int stat, lostr;

    lostr = strlen(response->string);
    GRQST (&wkid, &stdnr, &stat, &lostr, response->string);

    response->status = stat;
    response->string[lostr] = '\0';

    return gks_errno;
}
 

gcreateseg (segment_name)
    Gint segment_name;
{
    int segn = segment_name;

    GCRSG (&segn);

    return gks_errno;
}

 
gcopysegws (workstation_id, segment_name)
    Gint workstation_id, segment_name;
{
    int wkid = workstation_id, segn = segment_name;

    GCSGWK (&wkid, &segn);

    return gks_errno;
}
 

gredrawsegws (workstation_id)
    Gint workstation_id;
{
    int wkid = workstation_id;

    GRSGWK (&wkid);

    return gks_errno;
}
 

gcloseseg ()
{
    GCLSG ();

    return gks_errno;
}
 

gevaltran (ppoint, pshift, angle, pscale, coord, result)
    Gpoint *ppoint, *pshift;
    Gfloat angle;
    Gscale *pscale;
    Gcsw coord;
    Gfloat result[2][3];
{
    float x0, y0, tx, ty, phi, fx, fy;
    int isw, i, j;    
    float tran[2][3];

    x0  = ppoint->x;
    y0  = ppoint->y;
    tx  = pshift->x;
    ty  = pshift->y;
    phi = angle;
    fx  = pscale->x_scale;
    fy  = pscale->y_scale;
    isw = coord;

    GEVTM (&x0, &y0, &tx, &ty, &phi, &fx, &fy, &isw, tran);

    for (i = 0; i < 2; i++)
        for (j = 0; j < 3; j++)
            result[i][j] = tran[i][j];

    return gks_errno;
}


gsetsegtran (segment_name, segtran)
    Gint segment_name;
    Gfloat segtran[2][3];
{
    int segn = segment_name, i, j;
    float tran[2][3];

    for (i = 0; i < 2; i++)
        for (j = 0; j < 3; j++)
            tran[i][j] = segtran[i][j];

    GSSGT (&segn, tran);

    return gks_errno;
}


ginqopst (state)
    Gint *state;
{
    int opsta;

    GQOPS (&opsta);

    *state = opsta;

    return OK;
}
 

ginqlevelgks (level, error_status)
    Gint *level;
    Gint *error_status;
{
    int errind, lev;

    GQLVKS (&errind, &lev);

    *level = lev;
    *error_status = errind;

    return OK;
}
 

ginqmaxntrannum (maxtran, error_status)
    Gint *maxtran, *error_status;
{
    int errind, maxtnr;

    GQMNTN (&errind, &maxtnr);

    *maxtran = maxtnr;
    *error_status = errind;

    return OK;
}
 

/*
ginqopenws ()
{
} */


/*
ginqactivews ()
{
} */


ginqcharheight (height, error_status)
    Gint *error_status;
    Gfloat *height;
{
    int errind;
    float chh;

    GQCHH (&errind, &chh);

    *height = chh;
    *error_status = errind;

    return OK;
}
 

ginqcharup (up, error_status)
    Gpoint *up;
    Gint *error_status;
{
    float chux, chuy;
    int errind;

    GQCHUP (&errind, &chux, &chuy);

    up->x = chux;
    up->y = chuy;
    *error_status = errind;

    return OK;
}
 

ginqtextpath (text_path, error_status)
    Gtxpath *text_path;
    Gint *error_status;
{
    int errind, path;

    GQTXP (&errind, &path);

    *text_path = path;
    *error_status = errind;

    return OK;
}


ginqtextalign (align, error_status)
    Gtxalign *align;
    Gint *error_status;
{
    int errind, alh, alv;

    GQTXAL (&errind, &alh, &alv);

    align->hor = alh;
    align->ver = alv;

    return OK;
}

 
ginqasf (asflist, error_status)
    Gasfs *asflist;
    Gint *error_status;
{
    int errind, flag[13];

    GQASF (&errind, flag);

    asflist->ln_type    = flag[0];
    asflist->ln_width   = flag[1];
    asflist->ln_colour  = flag[2];
    asflist->mk_type    = flag[3];
    asflist->mk_size    = flag[4];
    asflist->mk_colour  = flag[5];
    asflist->tx_fp      = flag[6];
    asflist->tx_exp     = flag[7];
    asflist->tx_space   = flag[8];
    asflist->tx_colour  = flag[9];
    asflist->fl_inter   = flag[10];
    asflist->fl_style   = flag[11];
    asflist->fl_colour  = flag[12];
    *error_status = errind;

    return OK;
}
 

ginqlineind (index, error_status)
    Gint *index, *error_status;
{
    int errind, pli;

    GQPLI (&errind, &pli);

    *index = pli;
    *error_status = errind;

    return OK;
}


ginqlinetype (type, error_status)
    Gint *type;
    Gint *error_status;
{
    int errind, ltype;

    GQLN (&errind, &ltype);

    *type = ltype;
    *error_status = errind;

    return OK;
}


ginqlinewidth (width, error_status)
    Gfloat *width;
    Gint *error_status;
{
    int errind;
    float lwidth;

    GQLWSC (&errind, &lwidth);

    *width = lwidth;
    *error_status = errind;

    return OK;
}


ginqlinecolourind (index, error_status)
    Gint *index, *error_status;
{
    int errind, coli;

    GQPLCI (&errind, &coli);

    *index = coli;
    *error_status = errind;

    return OK;
}
 

ginqmarkerind (index, error_status)
    Gint *index, *error_status;
{
    int errind, pmi;

    GQPMI (&errind, &pmi);

    *index = pmi;
    *error_status = errind;

    return OK;
}


ginqmarkertype (level, error_status)
    Gint *level;
    Gint *error_status;
{
    int errind, mtype;

    GQMK (&errind, &mtype);

    *level = mtype;
    *error_status = errind;

    return OK;
}
 

ginqmarkersize (mksize, error_status)
    Gfloat *mksize;
    Gint *error_status;
{
    int errind;
    float mszsc;

    GQMKSC (&errind, &mszsc);

    *mksize = mszsc;
    *error_status = errind;

    return OK;
}
 

ginqmarkercolourind (index, error_status)
    Gint *index, *error_status;
{
    int errind, coli;

    GQPMCI (&errind, &coli);

    *index = coli;
    *error_status = errind;

    return OK;
}
 

ginqtextind (index, error_status)
    Gint *index, *error_status;
{
    int errind, txi;

    GQTXI (&errind, &txi);

    *index = txi;
    *error_status = errind;

    return OK;
}


ginqtextfontprec (txfp, error_status)
    Gtxfp *txfp;
    Gint  *error_status;
{
    int errind, font, prec;

    GQTXFP (&errind, &font, &prec);

    txfp->font = font;
    txfp->prec = prec;
    *error_status = errind;

    return OK;
}
 

ginqcharexpan (chexp, error_status)
    Gfloat *chexp;
    Gint *error_status;
{
    int errind;
    float chxp;

    GQCHXP (&errind, &chxp);

    *chexp = chxp;
    *error_status = errind;

    return OK;
}
 

ginqcharspace (chspc, error_status)
    Gfloat *chspc;
    Gint  *error_status;
{
    int errind;
    float chsp;

    GQCHSP (&errind, &chsp);

    *chspc = chsp;
    *error_status = errind;

    return OK;
}
 

ginqtextcolourind (index, error_status)
    Gint *index, *error_status;
{
    int errind, coli;

    GQTXCI (&errind, &coli);

    *index = coli;
    *error_status = errind;

    return OK;
}
 

ginqfillind (index, error_status)
    Gint *index, *error_status;
{
    int errind, fai;

    GQFAI (&errind, &fai);

    *index = fai;
    *error_status = errind;

    return OK;
}
 

ginqfillintstyle (style, error_status)
    Gint *style;
    Gint *error_status;
{
    int ints, errind;

    GQFAIS (&errind, &ints);

    *style = ints;
    *error_status = errind;

    return OK;
}
 

ginqfillstyle (index, error_status)
    Gint *index, *error_status;
{
    int styli, errind;

    GQFASI (&errind, &styli);

    *index = styli;
    *error_status = errind;

    return OK;
}
 

ginqfillcolourind (index, error_status)
    Gint *index, *error_status;
{
    int errind, coli;

    GQFACI (&errind, &coli);

    *index = coli;
    *error_status = errind;

    return OK;
}


ginqcurntrannum (tran, error_status)
    Gint *tran, *error_status;
{
    int errind, tnr;

    GQCNTN (&errind, &tnr);

    *tran = tnr;
    *error_status = errind;

    return OK;
}
 

ginqntran (num, tran, error_status)
    Gint num, *error_status;
    Gtran *tran;
{
    int tnr = num, errind;
    float wn[4], vp[4];

    GQNT (&tnr, &errind, wn, vp);

    tran->w.xmin = wn[0];
    tran->w.xmax = wn[1];
    tran->w.ymin = wn[2];
    tran->w.ymax = wn[3];
    tran->v.xmin = vp[0];
    tran->v.xmax = vp[1];
    tran->v.ymin = vp[2];
    tran->v.ymax = vp[3];
    *error_status = errind;

    return OK;
}
 

ginqclip (clipping, error_status)
    Gcliprect *clipping;
    Gint *error_status;
{
    int errind, clsw;
    float clrt[4];

    GQCLIP (&errind, &clsw, clrt);

    clipping->rec.xmin = clrt[0];
    clipping->rec.xmax = clrt[1];
    clipping->rec.ymin = clrt[2];
    clipping->rec.ymax = clrt[3];
    clipping->ind = clsw;
    *error_status = errind;

    return OK;
}
 

/*
ginqwsconntype ()
{
} */


ginqwscategory (workstation_type, cat, error_status)
    Gwstype *workstation_type;
    Gint *cat;
    Gint *error_status;
{
    int wstype = *workstation_type;
    int errind, category;

    GQWKCA (&wstype, &errind, &category);

    *cat = category;
    *error_status = errind;

    return OK;
}
 

ginqdisplaysize (workstation_type, dspsz, error_status)
    Gwstype *workstation_type;
    Gdspsize *dspsz;
    Gint *error_status;
{
    int wstype = *workstation_type;
    int errind, units, ras_x, ras_y;
    float px, py;
 
    GQDSP (&wstype, &errind, &units, &px, &py, &ras_x, &ras_y);

    dspsz->units = units;
    dspsz->device.x = px;
    dspsz->device.y = py;
    dspsz->raster.x = ras_x;
    dspsz->raster.y = ras_y;
    *error_status = errind;

    return OK;
}
 

ginqtextextent (workstation_id, position, string, extent, error_status)
    Gint workstation_id, *error_status;
    Gpoint *position;
    Gchar *string;
    Gextent *extent;
{
    int wkid = workstation_id, errind, nchars;
    float qx, qy, cpx, cpy, trx[4], try[4];
    char *chars;
#ifdef VMS
    struct dsc$descriptor_s text;
#endif
#ifdef cray
    _fcd text;
#endif

    qx = position->x;
    qy = position->y;
    chars = string;
    nchars = strlen(chars);
#ifdef VMS
    text.dsc$b_dtype = DSC$K_DTYPE_T;
    text.dsc$b_class = DSC$K_CLASS_S;
    text.dsc$w_length = nchars;
    text.dsc$a_pointer = chars;

    GQTXXS (&wkid, &qx, &qy, &nchars, &text, &errind, &cpx, &cpy, trx, try);
#else
#ifdef cray
    text = _cptofcd(chars, nchars);

    GQTXXS (&wkid, &qx, &qy, &nchars, text, &errind, &cpx, &cpy, trx, try);
#else
    GQTXXS (&wkid, &qx, &qy, &nchars, chars, &errind, &cpx, &cpy, trx, try,
        nchars);
#endif /* cray */
#endif /* VMS */

    extent->concat.x = cpx;
    extent->concat.y = cpy;
    extent->corner_1.x = trx[0];
    extent->corner_1.y = try[0];
    extent->corner_2.x = trx[1];
    extent->corner_2.y = try[1];
    extent->corner_3.x = trx[2];
    extent->corner_3.y = try[2];
    extent->corner_4.x = trx[3];
    extent->corner_4.y = try[3];

    return OK;
}
 

ginqnameopenseg (segment_name, error_status)
    Gint *segment_name;
    Gint *error_status;
{
    int errind, segn;

    GQOPSG (&errind, &segn);

    *segment_name = segn;
    *error_status = errind;

    return OK;
}


gemergencyclosegks ()
{
    GECLKS ();

    return OK;
}
